/*
    an advert skipper for 4:3 programs

    v0.01 - initial
    v0.02 - if program has adverts, auto skip the initial advert
          - detect pillar box 4:3 (e.g. on bbc4 old grey wt)
          - create /DataFiles/CutFiles directory if necessary
          - support for more firmwares (exTap_v0.13)
    v0.03 - prevent playback restarting when paused
          - cleaned up error handling
          - option to disable coloured keys
          - compiled with exTAP_v0.17, (support for firmware 12.88-13.59)

=================

    on playback started:
        check if cutfile exists
        if not scan the file being played
        if so, load the cutfile
        if aspect change, jump to next cutpoint

    on playback stopped:
        stop scanning the file - abandon cutfile if not complete

    on aspect change
        if ->16:9 try to jump to next 4:3

--- not in this release ---

on timeshift playback started:
    assume there is no cutfile, as it is likely to be old
    scan rec file from beginining
    keep scanning file as it may be growing
    if aspect change, jump to next cutpoint

on timeshift playback stopped:
    stop scanning the file

on recording playback started:
    assume there is no cutfile, as it is likely to be old
    scan rec file from beginining
    keep scanning file as it may be growing
    if aspect change, jump to next cutpoint

on recrding playback stopped:
    stop scanning the file

on channel change
    (should have just killed the timeshift; so timeshift playback will have changed)

on overnight scan
on demand
on really idle
    check there are no recordings running
    check there has been no activity for a while
    scan the disk
    if there if any activity, or playback change,  or recording change abort the scan

every 10 minutes
    update the cutfile list
    scan any new files first

--- --- --- --- ---
+todo:
seek is signed! - only works up to 2gb
-hack available
-need to fix for >4gb, need 64 bit fpos, or always use sectors

purge cutfiles on startup
remove timeshift cutfile
remove orphaned cutfiles

build list of .rec files
skip bbc
skip radio
skip those with valid (complete) cutfile
sort by newest first (most likely to be needed)
periodically detect if new file appears (e.g. via alttair, rename) + add to list
periodically detect if file disappears (e.g. via alttair, rename) + remove cutfile
i.e. periodically update todolist (30 mins)

allow scanning section to be used before it is complete
beware if scanning playfile that file is not closed, losing the hijacked connection

beware of hitting the end of the timeshift buffer; but don't stop; just throttle back
how to cope if the user save the timesfhift buffer (turns it intoa recording)
what to do if the user changes channel / trashes the timeshift buffer

*/
#include <string.h>
#include <tap.h>
#include "exTap.h"
#include "tf5800.h"
#include "ad_skipper_43.h"
#include "Drive.h"

#ifdef _WIN32
#else
#define ID_USERTAP 0xbdb01203
TAP_ID( ID_USERTAP );

TAP_PROGRAM_NAME("ad_skipper_43");
TAP_AUTHOR_NAME("bdb");
TAP_DESCRIPTION("v0.03 an advert skipper for 4:3 programs");
TAP_ETCINFO(__DATE__);
#endif

//-----------------------------------------------------------------------------
type_section section_playing = {0};

//----------------------------------------------------------------------------
int first_time = 1;
int automatic_skip_enabled = 1;
int coloured_keys_enabled = 1;

int two_seconds_enabled = 1;
dword two_seconds = 0;

int debug = 3;
type_scan pb_scan;

int last_tvRadio = SVC_TYPE_Tv;
int last_chNum = 0;
//----------------------------------------------------------------------------
//
//tsr_commander stuff
//
#ifdef _WIN32
#else
#include "TSRCommander.inc"

bool TSRCommanderExitTAP()
{
    close();
    return FALSE;
}

void TSRCommanderConfigDialog()
{
    ShowMenu(&main_menu);
}

int is_tsr_commander_running(void)
{
    int i;
    type_TAPProcess *TAPProcess;

    for (i=0; i<16; i++) {
        TAPProcess = (type_TAPProcess *)(fw_tap_table + (i*32));
        if (TAPProcess->firmwaregp) {
            if (TAPProcess->header) {
                if (TAPProcess->header->id == 0x807a10b2) {  //tsr_commander
                    return 1;
                }
                if (TAPProcess->header->id == 0x81001021) {  //tap_commander
                    return 1;
                }
            }
        }
    }
    return 0;
}
#endif
//----------------------------------------------------------------------------
//
//main
//
char *tap_name;
type_menu main_menu = {
    0,
    0,
    {0},
    0,
    (char *)__tap_program_name__,
    {
        {6, "xxx",              },
        {0,  "Manual Advert Skip",},
        {1,  "Rescan Current File",},
        {2,  "Disable Coloured Keys",},
        {3,  "Enable Serial Port Debug",},
        {4,  "Enable OSD Status",},

        {5,  "Exit TAP",},
    },
};

//
//-----------------------------------------------------------------------------
int TAP_Main(void)
{
    int ret;

    tap_name =  (char *)__tap_program_name__;
    sprintf(strLog, "Opening %s", tap_name);
    log_message(LOG_SCREEN_WAIT|LOG_SERIAL, strLog, DELAY_SHORT);
//    TAP_Print("\n_tap_startAddr=%x", _tap_startAddr);

    //determine support for different firmwares
    ret = exTAP_Init();
    if (ret) {
        sprintf(strLog, "Unsupported platform - %d:%04x", fw_sysid, _appl_version);
        log_message(LOG_SCREEN_WAIT|LOG_SERIAL, strLog, DELAY_MEDIUM);
        return 0;
    }
    if ((fw_fChangePlaybackPos == 0) || (fw_get_channel_info == 0)) {
        sprintf(strLog, "Unsupported platform - %d:%04x", fw_sysid, _appl_version);
        log_message(LOG_SCREEN_WAIT|LOG_SERIAL, strLog, DELAY_MEDIUM);
        return 0;
    }


    main_menu.visible = 0;
    pb_scan.state = STATE_IDLE;
    pb_scan.scanrate = 1000;

    if (automatic_skip_enabled) {
        strcpy(main_menu.item[1].str, "Manual Advert Skip");
    }
    else {
        strcpy(main_menu.item[1].str, "Automatic Advert Skip");
    }

    if (coloured_keys_enabled) {
        strcpy(main_menu.item[3].str, "Disable Coloured Keys");
    }
    else {
        strcpy(main_menu.item[3].str, "Enable Coloured Keys");
    }

    if (debug & 2) {
        strcpy(main_menu.item[4].str, "Disable Serial Port Debug");
    }
    else {
        strcpy(main_menu.item[4].str, "Enable Serial Port Debug");
    }

    if (debug & 1) {
        strcpy(main_menu.item[5].str, "Disable OSD Status");
    }
    else {
        strcpy(main_menu.item[5].str, "Enable OSD Status");
    }

    TAP_Channel_GetCurrent(&last_tvRadio, &last_chNum);
    TSRCommanderInit(0, TRUE);
    return 1;                 //0=application mode, 1=tsr mode
}

int close(void)
{
    DeleteMenu(&main_menu);

    sprintf(strLog, "Closing  %s", tap_name);
    log_message(LOG_SERIAL|LOG_SCREEN_WAIT, strLog, DELAY_SHORT);
    if (message_rgn) TAP_Osd_Delete(message_rgn);
    if (OpLog) TAP_Hdd_Fclose(OpLog);
    TSRCommanderErase();
    TAP_Print("\nClosing %s", tap_name);
    TAP_Exit();
    return 0;
}

dword TAP_EventHandler(word event, dword param1, dword param2)
{
    dword state, subState;
    int ret;
    int tick;
    static int next_tick = 0;
    TYPE_PlayInfo pi;        //overflows!
    byte dummy2[0x80];


    switch (event) {
        case EVT_IDLE:
//            TAP_Print("\nEVT_IDLE, event=%x, param1=%x, param2=%x", event, param1, param2);
            tick = TAP_GetTick();
            if (first_time) {
                init();
                first_time = 0;
            }

            //
            //message window
            //
            if (message_showing) {
                if (TAP_GetTick() > message_showing) {
                    TAP_Osd_Delete(message_rgn);
                    message_rgn = 0;
                    message_showing = 0;
                }
            }


            //
            //playback aspect ratio monitor
            //
            ret = exTAP_update_aspect_ratio();
            //if it has changed,
            if (ret&4) {
                if (debug & 2) {
                    TAP_Print("\nmain aspect changed ... ");
                    if      (fw_aspect[CHANNEL_Main].ratio==ASPECT_4_3)    TAP_Print("4:3");
                    else if (fw_aspect[CHANNEL_Main].ratio==ASPECT_16_9)   TAP_Print("16:9");
                    else if (fw_aspect[CHANNEL_Main].ratio==ASPECT_2_21_1) TAP_Print("2.21:1");
                    else                                                   TAP_Print("???");

                    TAP_Hdd_GetPlayInfo(&pi);
                    if (pi.playMode != PLAYMODE_None) {
                        TAP_Print("\nblock: %x/%x, fpos=%x", pi.currentBlock, pi.totalBlock, BYTESPERSECTOR*188*pi.currentBlock);
                    }
                }

                //if from 4:3 to 16:9
                if (automatic_skip_enabled) {
                    if (fw_aspect[CHANNEL_Main].ratio==ASPECT_16_9) {
                        skip_to_next(ASPECT_4_3, 0);
                    }
                }
            }
            //also detect pillarbox to 16:9, i.e. 16:9,afd:9,b,d --> 16:9,afd:8,a,e,f
            if (ret & 0x8) {
                //if from 4:3 to 16:9
                if (automatic_skip_enabled) {
                    if (fw_aspect[CHANNEL_Main].ratio==ASPECT_16_9) {
                        if ((fw_aspect[CHANNEL_Main].afd == 0x8) || (fw_aspect[CHANNEL_Main].afd == 0xa) || (fw_aspect[CHANNEL_Main].afd == 0xe) || (fw_aspect[CHANNEL_Main].afd == 0xf)) {
                            skip_to_next(ASPECT_4_3, 0);
                        }
                    }
                }
            }


            //if main changed
            if (ret & 0xc) if (debug & 2) display_aspect_ratio();



            //
            //housekeeping
            //
            if (two_seconds_enabled) {
                if (TAP_GetTick() > two_seconds) {
                    two_second_update();
                    two_seconds = TAP_GetTick() + (2*100);
                }
            }


            //
            //playback scan
            //try not to hammer the toppy too much
            //
//            if (tick > next_tick) {
//                playback_update(&pb_scan, &section_playing);
//                next_tick = tick + 300;             //every 3s, low
//            }
            if (next_tick) {
                next_tick --;
            }
            else {
                playback_update(&pb_scan, &section_playing);
                next_tick = pb_scan.scanrate;                   //every 100 idles
            }

            TSRCommanderWork();
            break;

        case EVT_KEY:
//            TAP_Print("\nEVT_KEY, event=%x, param1=%x, param2=%x", event, param1, param2);
            TAP_GetState(&state, &subState);
            if (main_menu.visible) {
                param1 = main_menu_handler(&main_menu, param1);
            }
            else
            if ((state == STATE_Normal) && (subState == SUBSTATE_Normal)) {
                param1 = key_handler(param1);
            }
            break;

        default:
//            TAP_Print("\nEVT_???, event=%x, param1=%x, param2=%x", event, param1, param2);
            break;
    }
    return param1;
}

int main_menu_handler(type_menu *menu, dword param1)
{
    int selection;

    switch (param1) {
        case RKEY_Ok:
            selection = TAP_Win_GetSelection(&menu->win);
            DeleteMenu(menu);
            switch (selection) {
                case 0: //Skip Mode
                    automatic_skip_enabled ^= 1;
                    if (automatic_skip_enabled) {
                        strcpy(menu->item[1].str, "Manual Advert Skip");
                    }
                    else {
                        strcpy(menu->item[1].str, "Automatic Advert Skip");
                    }
                    return 0;
                case 1: //rescan current file now
                    restart_scan();
                    return 0;
                case 2: //coloured keys
                    coloured_keys_enabled ^= 1;
                    if (coloured_keys_enabled) {
                        strcpy(menu->item[3].str, "Disable Coloured Keys");
                    }
                    else {
                        strcpy(menu->item[3].str, "Enable Coloured Keys");
                    }
                    return 0;
                case 3: //serial debug
                    debug ^= 2;
                    if (debug & 2) {
                        strcpy(menu->item[4].str, "Disable Serial Port Debug");
                    }
                    else {
                        strcpy(menu->item[4].str, "Enable Serial Port Debug");
                    }
                    return 0;
                case 4: //osd debug
                    debug ^= 1;
                    if (debug & 1) {
                        strcpy(menu->item[5].str, "Disable OSD Status");
                    }
                    else {
                        strcpy(menu->item[5].str, "Enable OSD Status");
                    }
                    return 0;

                case 5: //exit
                    close();
                    return 0;
            }
            break;
        case RKEY_Recall:
            DeleteMenu(menu);
            return param1; // cascades key to other TAP's
        case RKEY_Exit:
            DeleteMenu(menu);
            return 0; // exit without Cascading
        default:
            TAP_Win_Action(&menu->win, param1); // send all other key's to menu-Win
            return 0;
    }
    return param1;
}

int key_handler(dword param1)
{
    switch(param1) {
        //only if tsrcommander is not running
        case RKEY_Recall:
            if (!is_tsr_commander_running()) {
                ShowMenu(&main_menu);
                return 0;
            }
            break;

       case RKEY_Red:                       //status
            if (coloured_keys_enabled) {
                restart_scan();
                return 0;
            }
            break;
        case RKEY_Green:                    //go start of adverts
            if (coloured_keys_enabled) {
                skip_to_next(ASPECT_16_9, 0);
                return 0;
            }
            break;
        case RKEY_Yellow:                   //go back to end of adverts
            if (coloured_keys_enabled) {
                skip_to_next(ASPECT_4_3, 1);
                return 0;
            }
            break;
        case RKEY_Blue:                     //go end of adverts
            if (coloured_keys_enabled) {
                skip_to_next(ASPECT_4_3, 0);
                return 0;
            }
            break;
    }
    return param1;
}

int restart_scan()
{
    TYPE_PlayInfo pi;        //overflows!
    byte dummy2[0x80];
    char cut_filename[0x100];
    int len;
    int ret;

    TAP_Hdd_GetPlayInfo(&pi);
    if (debug & 2) TAP_Print("\n Play Mode: %s", get_str(playmode_str, pi.playMode));
    if ((pi.playMode == PLAYMODE_Playing) || (pi.playMode == PLAYMODE_RecPlaying)) {
        if (pb_scan.state == STATE_IDLE) {
            if (debug & 2) TAP_Print("\nrestarting scan");

            //create .cut filename
            len = strlen(pi.file->name);
            if (strncmp(&pi.file->name[len-4], ".rec", 4) != 0) {
                //if not playing a .rec file, quit
                if (debug & 2) TAP_Print("\nnot playing a .rec file, aborting");
                return 0;
            }
            memcpy(&cut_filename[0], pi.file->name, len-4);
            memcpy(&cut_filename[len-4], ".cut\0", 5);

            //delete the .cut file
            TAP_Hdd_ChangeDir("/");
            ret = TAP_Hdd_ChangeDir("DataFiles");
            if (ret == 0) {
                return 0;
            }
            ret = TAP_Hdd_ChangeDir("CutFiles");
            if (ret == 0) {
                return 0;
            }
            if (TAP_Hdd_Exist(cut_filename)) {
                if (debug & 2) TAP_Print("\n%s exists, delete it", cut_filename);
                TAP_Hdd_Delete(cut_filename);
            }

            section_playing.info.scanned = 0;
            section_playing.aspect.count = 0;
            section_playing.iframe.count = 0;
            section_playing.gframe.count = 0;

            //restart the scan
//            pb_scan.flags = 0;
            pb_scan.state = STATE_OPEN;
        }
        else {
            if (debug & 2) TAP_Print("\nstopping scan");
            pb_scan.scanrate = 1000;
            pb_scan.abort = 1;
            pb_scan.state = STATE_CLOSE;
        }
    }
    return 0;
}

//-----------------------------------------------------------------------------
void init()
{
#ifdef _WIN32
#else
    Hdd_Init();
#endif
}





int display_aspect_ratio()
{
    int j;
    int main_sub;

    for (j=0; j<2; j++) {
        main_sub = (j==0) ? CHANNEL_Main : CHANNEL_Sub;

        TAP_Print("\n%s Channel", (main_sub == CHANNEL_Sub) ? " Sub" : "Main");
        TAP_Print("        ");
        TAP_Print("%dx%d, ", fw_aspect[main_sub].horizontal_size, fw_aspect[main_sub].vertical_size);
        if      (fw_aspect[main_sub].ratio==ASPECT_4_3)    TAP_Print(" 4:3");
        else if (fw_aspect[main_sub].ratio==ASPECT_16_9)   TAP_Print("16:9");
        else if (fw_aspect[main_sub].ratio==ASPECT_2_21_1) TAP_Print("2.21:1");
        else                                               TAP_Print("??:??");
        TAP_Print(", AFD_%x ", fw_aspect[main_sub].afd);
        if (fw_aspect[main_sub].ratio == 2) {
            switch(fw_aspect[main_sub].afd) {
                case AFD0: TAP_Print("4:3 full frame image                          "); break;
                case AFD1: TAP_Print("4:3 full frame image                          "); break;
                case AFD2: TAP_Print("16:9 letterbox image                          "); break;
                case AFD3: TAP_Print("14:9 letterbox image                          "); break;
                case AFD5: TAP_Print("4:3 full frame image, alternative 14:9 center "); break;
                case AFD6: TAP_Print("16:9 letterbox image, alternative 14:9 center "); break;
                case AFD7: TAP_Print("16:9 letterbox image, alternative 4:3 center  "); break;
                default  : TAP_Print("Reserved                                      "); break;
            }
        }
        else if (fw_aspect[main_sub].ratio == 3) {
            switch(fw_aspect[main_sub].afd) {
                case AFD0: TAP_Print("16:9 full frame image                          "); break;
                case AFD1: TAP_Print("4:3 pillarbox image                            "); break;
                case AFD2: TAP_Print("16:9 full frame image                          "); break;
                case AFD3: TAP_Print("14:9 pillarbox image                           "); break;
                case AFD5: TAP_Print("4:3 pillarbox image, alternative 14:9 center   "); break;
                case AFD6: TAP_Print("16:9 full frame image, alternative 14:9 center "); break;
                case AFD7: TAP_Print("16:9 full frame image, alternative 4:3 center  "); break;
                default  : TAP_Print("Reserved                                       "); break;
            }
        }
    }
}
