
#include <string.h>
#include <tap.h>
#include "exTap.h"
#include "tf5800.h"
#include "ad_skipper_43.h"
#include "Drive.h"

//-----------------------------------------------------------------------------
int scan_ts_for_aspect(ulong64 foffset64, type_section *section, type_scan *scan)
{
    int j,k;
    int pus;
    int pid;
    int af;
    int af_offset;
    int aspect_ratio_information;
    ulong64 offset64;

//    TAP_Print("\nscan_ts_for_aspect(), foffset64=%x%08x, ts_len=%x", *(dword *)(0+(int)&foffset64), *(dword *)(4+(int)&foffset64), scan->ts_len);

    //are we on a ts boundary?
    j = 0;
    while ((scan->ts_buf[j] != 0x47) || (scan->ts_buf[j+0xbc] != 0x47) || (scan->ts_buf[j+0xbc+0xbc] != 0x47)) {
        j++;
        //got to end of buffer
        if (j > (int)(scan->ts_len-0xbc-0xbc)) {
            if (debug & 2) {
                TAP_Print("\nare we on a ts boundary?");
                dump((dword)&scan->ts_buf[j], 0xbc);
            }
            return j;
        }
    }

    while (j<=(int)(scan->ts_len-0xbc)) {
        if  (scan->ts_buf[j] != 0x47) {
            if (debug & 2) TAP_Print("\noops - bad flag, j=%x", j);
            return j;
        }
        pid = ((scan->ts_buf[j+1]<<8) + scan->ts_buf[j+2]) & 0x1fff;

        //only interested in video pid
        if ((scan->video_pid == 0) || (pid == scan->video_pid)) {

            //ignore the adaptation field
            af = (scan->ts_buf[j+3]>>4) & 0x3;
            af_offset = 0;
            if (af >= 2) {
                af_offset = scan->ts_buf[j+4] + 1;
            }

            //start of pes packets, will have the pus set
            pus = (scan->ts_buf[j+1] >> 6) & 0x1;
            if (pus) {
                //is it a pes packet?
                if ((scan->ts_buf[j+4+af_offset+0] == 0x00) && (scan->ts_buf[j+4+af_offset+1] == 0x00) && (scan->ts_buf[j+4+af_offset+2] == 0x01)) {
//                    TAP_Print("\npes: %08x : 00 00 01 %02x", j, scan->ts_buf[j+4+af_offset+3]);

                    //look for picture_start_code
                    k = 4+af_offset+4;
                    while (k<(0xbc-2)) {
                        if ((scan->ts_buf[j+k+0] == 0x00) && (scan->ts_buf[j+k+1] == 0x00) && (scan->ts_buf[j+k+2] == 0x01)) {
//                            TAP_Print("\npes: %08x : 00 00 01 %02x : ", j, scan->ts_buf[j+k+3]);

                            switch(scan->ts_buf[j+k+3]) {
//                                case 0x00: //picture_start_code
//                                    TAP_Print("\npicture_start_code:   ");
//                                    break;

                                case 0xb2: //user_data_start_code
//                                    TAP_Print("\nuser_data_start_code: ");
//Active_format   4:3 coded frames                                16:9 coded frames
//0000  0001 Reserved                                        Reserved
//0010  0100 Reserved                                        Reserved
//0101  0111 Reserved                                        Reserved
//1000          4:3 full frame image                            16:9 full frame image
//1001          4:3 full frame image                            4:3 pillarbox image
//1010          16:9 letterbox image                            16:9 full frame image
//1011          14:9 letterbox image                            14:9 pillarbox image
//1100          Reserved                                        Reserved
//1101          4:3 full frame image, alternative 14:9 center   4:3 pillarbox image, alternative 14:9 center
//1110          16:9 letterbox image, alternative 14:9 center   16:9 full frame image, alternative 14:9 center
//1111          16:9 letterbox image, alternative 4:3 center    16:9 full frame image, alternative 4:3 center

                                    if ((scan->ts_buf[j+k+4] == 'D') && (scan->ts_buf[j+k+5] == 'T') &&
                                        (scan->ts_buf[j+k+6] == 'G') && (scan->ts_buf[j+k+7] == '1') &&
                                        (scan->ts_buf[j+k+8] == 'A')){
                                        scan->afd = scan->ts_buf[j+k+9]&0xf;
//                                        TAP_Print("\nafd: %x", scan->afd);
                                    }
                                    break;

                                case 0xb3: //sequence_header_code
//                                    TAP_Print("\nsequence_header_code: ");

                                    aspect_ratio_information = scan->ts_buf[j+k+7]>>4;

                                    if (aspect_ratio_information == ASPECT_16_9) {
                                        //pillar box 4:3 in 16:9 frame
                                        if (scan->afd == 0x9) {
                                            aspect_ratio_information = ASPECT_4_3;
                                        }
                                        //pillar box 14:9 in 16:9 frame
                                        if (scan->afd == 0xb) {
                                            aspect_ratio_information = ASPECT_4_3;
                                        }
                                        //pillar box 4:3 in 16:9 frame
                                        if (scan->afd == 0xd) {
                                            aspect_ratio_information = ASPECT_4_3;
                                        }
                                    }
//                                    TAP_Print("aspect: %s", (aspect_ratio_information==ASPECT_4_3) ? "4:3" : (aspect_ratio_information==ASPECT_16_9) ? "16:9" : "??:?");

                                    //has aspect ratio changed?
                                    if (aspect_ratio_information != scan->aspect_ratio_information) {

                                        if (scan->fine == 0) {
                                            //if in course mode, quit now, got a change
                                            //but if this is perfect, we dont want to go back... could end up in a loop
                                            return -1;
                                        }

                                        if (section->aspect.count < MAX_CUTPOINTS) {
                                            section->aspect.cut[section->aspect.count].from   = scan->aspect_ratio_information;
                                            section->aspect.cut[section->aspect.count].to     = aspect_ratio_information;
                                            section->aspect.cut[section->aspect.count].fpos64 = foffset64+j;
                                            section->aspect.count++;
                                        }
                                        offset64 = foffset64+j;
                                        if (debug & 2) {
//                                            TAP_Print("\n***");
                                            TAP_Print("\naspect change: from %s to %s, foffset64 = %x%08x",
                                                (scan->aspect_ratio_information==ASPECT_4_3) ? "4:3" : (scan->aspect_ratio_information==ASPECT_16_9) ? "16:9" : "??:?",
                                                (aspect_ratio_information==ASPECT_4_3) ? "4:3" : (aspect_ratio_information==ASPECT_16_9) ? "16:9" : "??:?",
                                                *(dword *)(0+(int)&offset64),
                                                *(dword *)(4+(int)&offset64)
                                            );
//                                            TAP_Print("\n***");
                                        }
                                        scan->aspect_ratio_information = aspect_ratio_information;
                                        if (scan->fine == 1) {
                                           //all done
                                            return -2;
                                        }
                                    }

                                    //stop as soon as we find one
                                    if (scan->fine == 1) {
                                    }
                                    else {
                                        return -3;
                                    }
                                    break;

//                                case 0xb5: //extension_start_code
//                                    TAP_Print("\nextension_start_code: ");
//                                    break;
//                                case 0xb7: //sequence_end_code
//                                    TAP_Print("\nsequence_end_code:    ");
//                                    break;
//                                case 0xb8: //group_start_code
//                                    TAP_Print("\ngroup_start_code:     ");
//                                    break;
//                                default:
//                                    //01-af = slice_start_code
//                                    //b9-ff = system_start_code
//                                    break;
                            }
                        }
                        k++;
                    }
                }
                else {
                    //could be encrypted, or just corrupted
//                    TAP_Print("\npus; not a pes");
//                    dump((dword)&scan->ts_buf[j], 0xbc);
                }
            }
        }
        j += 0xbc;
    }
    return j;
}

//-----------------------------------------------------------------------------
int scan_init(type_scan *scan, type_section *section, char *cut_filename)
{

    return 0;
}

int scan_open(type_scan *scan, type_section *section, char *cut_filename)
{
    int ret;
    int len;
    type_rec_header rec;
    type_channel_info channel_info;
    int chType, chNum;
    TYPE_PlayInfo pi;        //overflows!
    byte dummy2[0x80];


    ret = load_cutfile(cut_filename, section);
//-1=error
//1=.cut file does not exist
//0=ok/already loaded
    if (ret < 0) {
        if (debug & 2) TAP_Print("\nload_cutfile(), error, aborting");
        return -1;
//        scan->state = STATE_IDLE;
//        break;
    }
    if (ret == 1) {
        if (debug & 2) TAP_Print("\ncutfile does not exist");
    }
    if (debug & 2) dump_section(section);

    //get playing file info
    TAP_Hdd_GetPlayInfo(&pi);

    scan->scanrate = 1000;
    scan->playmode = pi.playMode;
    scan->fine = 0;
    scan->flags = 0;
    scan->aspect_ratio_information = 0;
    scan->afd = 0;
    scan->fpos64 = 0;
    scan->flen64 = 0;
    scan->flags = section->info.analysed;

    osd_status(section, scan);


    //if the cutfile exists use it
    //else scan for another
    //don't ignore if the cutfile is not complete
    if ((ret == 0) && (section->info.scanned == 1)) {
//        TAP_Print("\nload_cutfile(), cutfile loaded, no need to scan");
        //kick skip_to_next, in case we are in trailer
        on_playback_aspect_change();
        return 1;
    }


    section->info.loaded = 1;
    section->info.scanning = 1;
    section->info.scanned = 0;
    section->info.analysed = 0;

    //open the .rec file; well just hijack the already playing file
    //can we open the file while it is playing? - yes
    if (scan->playmode == PLAYMODE_None) {    //no playmode
        //open the .rec file
        if (debug & 2) TAP_Print("\nReading %s ...", section->info.filename);
        ret = TAP_Hdd_ChangeDir("/");
        ret = TAP_Hdd_ChangeDir("DataFiles");
        scan->Ip = TAP_Hdd_Fopen(section->info.filename);
        if (!scan->Ip) {
            if (debug & 2) TAP_Print("\nERROR: advert_scan(), %s not found", section->info.filename);
            scan->state = STATE_IDLE;
            return -1;
        }
    }
    else {
        scan->Ip = pi.file;       //hijack playing file
    }
#ifdef _WIN32
#else
    ret = Hdd_Fopen(scan->Ip, 1);
#endif
#ifdef _WIN32
    scan->flen64 = TAP_Hdd_Flen(scan->Ip);
    TAP_Print("\nflen = %x", scan->flen64);
#else
//    scan->flen = *(dword *)(4+(int)&scan->Ip->size);      //64bit???!!
//    TAP_Print("\nflen = %x", scan->flen);

    scan->flen64 = scan->Ip->size;                        //64bit???!!
    if (debug & 2) TAP_Print("\nflen64 = %x%08x", *(dword *)(0+(int)&scan->flen64), *(dword *)(4+(int)&scan->flen64));
#endif

//for playing file ?
    if ((scan->playmode == PLAYMODE_TempPlaying) || (scan->playmode == PLAYMODE_RecPlaying)){
//        scan->flen = TAP_Hdd_Flen(pi.file);       //will this work on an open file? -use Ip.size?
//        TAP_Print("\nflen              =%x", scan->flen);
//
        if (debug & 2) TAP_Print("\nflen = 512*188*%x=%x", pi.totalBlock, BYTESPERSECTOR*188*pi.totalBlock);
        scan->flen64 = BYTESPERSECTOR*188*pi.totalBlock;
        if (debug & 2) TAP_Print("\nflen64 = %x%08x", *(dword *)(0+(int)&scan->flen64), *(dword *)(4+(int)&scan->flen64));
    }


    //read the .rec header
#ifdef _WIN32
    TAP_Hdd_Fseek(scan->Ip, 0, SEEK_SET);
    len = TAP_Hdd_Fread(&rec, 0x200, 1, scan->Ip);
#else
    Hdd_Fseek(0);
    len = Hdd_Fread(0x200, (byte *)&rec);
#endif

    //for playing file
    if (rec.magic != 0x54467263) {    //TFrc
        chType = SVC_TYPE_Tv;
        chNum = 0xffff;
        if (debug & 2) TAP_Print("\nfw_get_channel_info(%x, %x, %x)", chType, chNum, &channel_info);
        ret = exTAP_CallFirmware(fw_get_channel_info, chType, chNum, (dword)&channel_info, 0,0,0,0,0,0,0,0);
        if (debug & 2) TAP_Print(", ret=%x", ret);
        if (debug & 2) TAP_Print("\nvideo_pid = %04x", channel_info.video_pid);
    }
    else {
        if (debug & 2) dump((int)&rec, 0x80);
        scan->video_pid = to_be16(rec.video_pid);
        if (debug & 2) TAP_Print("\nvideo_pid = %04x", scan->video_pid);
    }



    //start scanning
    scan->ts_buf = (byte *)malloc(TS_BUF_LEN+0x200);        //1Mbyte + a sector
    scan->ts_len = 0;
    scan->twrite = 0;
    scan->tread = 0;
    scan->fpos64 = 0x1000;       //skip .rec header  (but stick to sector boundary)
    scan->flen64 -= scan->fpos64;

    return 0;
}

int scan_update(type_scan *scan, type_section *section)
{
    int ret;
    ulong64 foffset64;
    int pos, len;

#ifdef _WIN32
    TAP_Hdd_Fseek(scan->Ip, scan->fpos64, SEEK_SET);
    ret = TAP_Hdd_Fread(&scan->ts_buf[scan->twrite], TS_BUF_LEN, 1, scan->Ip);
#else
    Hdd_Fseek(scan->fpos64>>9);
    ret = Hdd_Fread(TS_BUF_LEN, &scan->ts_buf[scan->twrite]);  //reads whole sectors (0x200)
#endif


    scan->ts_len = scan->twrite + TS_BUF_LEN;
    foffset64 = scan->fpos64 - scan->twrite;
if (0) {
TAP_Print("\nret=%x", ret);
TAP_Print(", fpos64=%x%08x", *(dword *)(0+(int)&scan->fpos64), *(dword *)(4+(int)&scan->fpos64));
TAP_Print(", tread=%x", scan->tread);
TAP_Print(", twrite=%x", scan->twrite);
TAP_Print(", TS_BUF_LEN=%x", TS_BUF_LEN);
TAP_Print(", ts_len=%x", scan->ts_len);
TAP_Print(", foffset=%x%08x", *(dword *)(0+(dword)&foffset64), *(dword *)(4+(dword)&foffset64));
}

    scan->tread = scan_ts_for_aspect(foffset64, section, scan);

    if (debug & 2) TAP_Print("\n%s scan, %x%08x/%x%08x", (scan->fine)?"  fine":"course", *(dword *)(0+(int)&scan->fpos64), *(dword *)(4+(int)&scan->fpos64), *(dword *)(0+(int)&scan->flen64), *(dword *)(4+(int)&scan->flen64));
    if (debug & 2) TAP_Print("%s",
        ((scan->flags & 3) == 1) ? ", [4:3]" :
        ((scan->flags & 3) == 2) ? ", [16:9]" :
        ((scan->flags & 3) == 3) ? ", [4:3/16:9]" : "");
    pos = scan->fpos64>>16;
    len = scan->flen64>>16;
    if (debug & 2) TAP_Print(", %d%%", (pos*100)/(len+1));
//pos =
//-1 = got a change, so go back, and switch to fine
//-2 = got a change, switch to course
//-3 = terminated, no change

    //switch back to course mode
    if (scan->tread == -2) {
        scan->fine = 0;
        scan->tread = -3;
        if (section->aspect.count) {
            if (section->aspect.cut[section->aspect.count-1].to == ASPECT_4_3)  scan->flags |= 1;
            if (section->aspect.cut[section->aspect.count-1].to == ASPECT_16_9) scan->flags |= 2;
        }
    }

    //found a aspect, but no change, move to the start of the next block
    if (scan->tread == -3) {
        scan->fpos64 += TS_BUF_LEN;
        if (scan->fine == 0) scan->fpos64 += (0xc00000-TS_BUF_LEN);   //go forward ~30s  30s@2Mbps=7.5M         30s@1.5Mbps=5.6M
        scan->twrite = 0;
    }

    //go back a block, set fine mode
    if (scan->tread == -1) {
        if (scan->fpos64 >= 0xc00000) {
            scan->fpos64 -= 0xc00000;
        } else {
            scan->fpos64 = 0x1000;
        }
        scan->fine = 1;
        scan->twrite = 0;
    }


    //else continue from here
    if (scan->tread >= 0) {
        scan->fpos64 += TS_BUF_LEN;
        scan->twrite = scan->ts_len - scan->tread;
        memcpy(&scan->ts_buf[0], &scan->ts_buf[scan->tread], scan->twrite);
    }


    //got to end of file?
    //what happens if file is growing? - shouldn't be in scanning mode
    if ((scan->flen64 - scan->fpos64) <= 0xc00000) {
        return 1;
    }
    if (scan->fpos64 > scan->flen64) {
        if (debug & 2) TAP_Print("\n*** oops ***");
        return 1;
    }
    return 0;
}

int scan_close(type_scan *scan, type_section *section, char *cut_filename)
{
    int i;

    //close files
    //don;t close file if we hijacked the playing one
    if (scan->playmode != PLAYMODE_None) {
#ifdef _WIN32
        TAP_Hdd_Fclose(scan->Ip);
#else
        Hdd_Fclose();
#endif
    }
    free(scan->ts_buf);

    //save the cut file
    section->info.loaded = 1;
    section->info.scanning = 0;
    section->info.scanned = (scan->abort) ? 0 : 1;
    for (i=0; i<(int)section->aspect.count; i++) {
        if (section->aspect.cut[i].to == ASPECT_4_3) section->info.analysed |= 1;
        if (section->aspect.cut[i].to == ASPECT_16_9) section->info.analysed |= 2;
    }
    if (debug & 2) dump_section(section);
    osd_status(section, scan);

    if (scan->abort == 0) {
        if (debug & 2) TAP_Print("\nsave_cutfile(), cut_filename=%s", cut_filename);
        save_cutfile(section, cut_filename);
    }

    return 0;
}

//-----------------------------------------------------------------------------
//scan the playing file
//
int playback_update(type_scan *scan, type_section *section)
{
    int ret;
    TYPE_PlayInfo pi;        //overflows!
    byte dummy2[0x80];
    static char cut_filename[0x100];
    int len;
    dword etime;


//    TAP_Print("\nplayback_update()");
    switch (scan->state) {
        case STATE_IDLE:
            break;
        case STATE_OPEN:
            if (debug & 2) TAP_Print("\nPB STATE_OPEN");

            //get playing file info
            TAP_Hdd_GetPlayInfo(&pi);
            if (pi.playMode == PLAYMODE_None) {
                //not playing file
                scan->state = STATE_IDLE;
                break;
            }
            if (debug & 2) {
//                TAP_Print("\nplayMode.file = %x", pi.file);
                TAP_Print("\nplayMode.name = %s", pi.file->name);
                dumpPlaying();
//                dump((int)pi.file, 0x90);
            }
//            last_playMode = pi.playMode;


            //open the cutfile
            len = strlen(pi.file->name);
            if (strncmp(&pi.file->name[len-4], ".rec", 4) != 0) {
                //if not playing a .rec file, quit
                if (debug & 2) TAP_Print("\nnot playing a .rec file, aborting");
                scan->state = STATE_IDLE;
                break;
            }

            //create .cut filename
            memcpy(&cut_filename[0], pi.file->name, len-4);
            memcpy(&cut_filename[len-4], ".cut\0", 5);
//            TAP_Print("\npi.file->name = %s, len=%x", pi.file->name, len);
            if (debug & 2) TAP_Print("\ncut_filename = %s", cut_filename);

            ret = scan_open(scan, section, cut_filename);
            //if abort
            if (ret < 0) {
                if (debug & 2) TAP_Print("\nERROR: scan_open(), %s", section->info.filename);
                scan->state = STATE_IDLE;
                return -1;
            }
            //if the cutfile exists use it
            //else scan for another
            //don't ignore if the cutfile is not complete
            if (ret == 1) {
                if (debug & 2) TAP_Print("\nload_cutfile(), cutfile loaded, no need to scan");
                //kick skip_to_next, in case we are in trailer
                on_playback_aspect_change();
                scan->state = STATE_IDLE;
                break;
            }

            scan->stime = TAP_GetTick();
            scan->state = STATE_UPDATE;
            break;

        case STATE_UPDATE:
//            TAP_Print("\nPB STATE_UPDATE");

            ret = scan_update(scan, section);
            if (ret == 1) {
                if ((scan->playmode == PLAYMODE_TempPlaying) || (scan->playmode == PLAYMODE_RecPlaying)) {
                    //could just go back a bit more
                    //and slow down just go away for a while
                    scan->fpos64 = scan->flen64 - 0xc00000;
                    scan->scanrate = 100000;             //may want to reduce this sometimes?
                }
                else {
                  scan->state = STATE_CLOSE;
                }
            }
            break;

        case STATE_CLOSE:
            if (debug & 2) TAP_Print("\nPB STATE_CLOSE");

            ret = scan_close(scan, section, cut_filename);

            etime = TAP_GetTick();
            if (debug & 2) TAP_Print("\ntook %d seconds", (etime-scan->stime)/100);

            //kick skip_to_next, in case we are in trailer
            on_playback_aspect_change();

            scan->abort = 0;
            scan->state = STATE_IDLE;
            break;
    }
    return 0;
}

//-----------------------------------------------------------------------------
